from django.urls import path
from .api import (
    TransitionApiView ,
    ActionListApi,
    DetailsListApiView,
    WorkFlowitemsListApi,
    WorkEventsListApi
)


urlpatterns = [
    path('action/',ActionListApi.as_view()),
    path('model/',DetailsListApiView.as_view()),
    path('workflowitems/',WorkFlowitemsListApi.as_view()),
    path('workevents/',WorkEventsListApi.as_view()),
    path('transition/',TransitionApiView.as_view())
]



from django.db.models import TextChoices
from django.utils.translation import gettext_lazy as _



class StateChoices(TextChoices):
    STATUS_DRAFT = 'DRAFT', _('DRAFT')
    STATUS_AW_APPROVAL = 'AWAITING_APPROVAL', _('AWAITING_APPROVAL')
    STATUS_AW_ACCEPT = 'AWAITING_ACCEPTANCE', _('AWAITING_ACCEPTANCE')
    STATUS_ACCEPTED = 'ACCEPTED', _('ACCEPTED')
    STATUS_APPROVED = 'APPROVED', _('APPROVED')
    STATUS_REJECTED = 'REJECTED', _('REJECTED')
    STATUS_FINANCE_REQUESTED = 'FINANCE_REQUESTED', _('FINANCE_REQUESTED')
    STATUS_COMPLETED = 'COMPLETED', _('COMPLETED')
    STATUS_AWAITING_BUYER_APPROVAL = 'AWAITING_BUYER_APPROVAL' , _('AWAITING_BUYER_APPROVAL')
    STATUS_APPROVED_BY_BUYER = 'APPROVED_BY_BUYER' , _('APPROVED_BY_BUYER')
    STATUS_REJECTED_BY_BUYER = 'REJECTED_BY_BUYER' , _('REJECTED_BY_BUYER')
    STATUS_ARCHIVED = 'STATUS_ARCHIVED' , _('STATUS_ARCHIVED')
    STATUS_FINANCED = 'FINANCED', _('FINANCED')
    STATUS_FINANCE_REJECTED = 'FINANCE_REJECTED', _('FINANCE_REJECTED')
    STATUS_SETTLED = 'SETTLED', _('SETTLED')
    STATUS_OVERDUE = 'OVERDUE', _('OVERDUE')
    STATUS_NO_AWAITING_SIGN = 'NO_AWAITING_SIGN', _('NO_AWAITING_SIGN')
    STATUS_AWAITING_SIGN_A = 'AWAITING_SIGN_A', _('AWAITING_SIGN_A')
    STATUS_AWAITING_SIGN_B = 'AWAITING_SIGN_B', _('AWAITING_SIGN_B')
    STATUS_AWAITING_SIGN_C = 'AWAITING_SIGN_C', _('AWAITING_SIGN_C')
    STATUS_AWAITING_SIGN_D = 'AWAITING_SIGN_D', _('AWAITING_SIGN_C')
    STATUS_AWAITING_SIGN_E = 'AWAITING_SIGN_E', _('AWAITING_SIGN_E')
    STATUS_AWAITING_SIGN_F = 'AWAITING_SIGN_F', _('AWAITING_SIGN_F')
    STATUS_AWAITING_SIGN_G = 'AWAITING_SIGN_G', _('AWAITING_SIGN_G')
    STATUS_AWAITING_SIGN_H = 'AWAITING_SIGN_H', _('AWAITING_SIGN_H')
    STATUS_AWAITING_SIGN_I = 'AWAITING_SIGN_I', _('AWAITING_SIGN_I')
    STATUS_AWAITING_SIGN_J = 'AWAITING_SIGN_J', _('AWAITING_SIGN_J')
    STATUS_AWAITING_SIGN_K = 'AWAITING_SIGN_K', _('AWAITING_SIGN_K')
    STATUS_AWAITING_SIGN_L = 'AWAITING_SIGN_L', _('AWAITING_SIGN_L')
    STATUS_AWAITING_SIGN_M = 'AWAITING_SIGN_M', _('AWAITING_SIGN_M')
    STATUS_AWAITING_SIGN_N = 'AWAITING_SIGN_N', _('AWAITING_SIGN_N')
    STATUS_AWAITING_SIGN_O = 'AWAITING_SIGN_O', _('AWAITING_SIGN_O')
    STATUS_AWAITING_SIGN_P = 'AWAITING_SIGN_P', _('AWAITING_SIGN_P')
    STATUS_AWAITING_SIGN_Q = 'AWAITING_SIGN_Q', _('AWAITING_SIGN_Q')
    STATUS_AWAITING_SIGN_R = 'AWAITING_SIGN_R', _('AWAITING_SIGN_R')
    STATUS_AWAITING_SIGN_S = 'AWAITING_SIGN_S', _('AWAITING_SIGN_S')
    STATUS_AWAITING_SIGN_T = 'AWAITING_SIGN_T', _('AWAITING_SIGN_T')
    STATUS_DELETED = 'DELETED', _('DELETED')
    # action and subaction
    SIGN_A = 'SIGN_A',_('SIGN_A')
    APPROVE = 'APPROVE',_('APPROVE')
    ACCEPT = 'ACCEPT',_('ACCEPT')
    ARCHIVE = 'ARCHIVE',_('ARCHIVE')
    SETTLE = 'SETTLE',_('SETTLE')
    REJECT = 'REJECT',_('REJECT')
    RETURN = 'RETURN',_('RETURN')
    SIGN_B = 'SIGN_B',_('SIGN_B')
    SIGN_C = 'SIGN_C',_('SIGN_C')
    REQUEST_FINANCE = 'REQUEST_FINANCE',_('REQUEST_FINANCE')
    SUBMIT = 'SUBMIT',_('SUBMIT')
    MAKER = 'MAKER',_('MAKER')
    INITIAL_STATE = 'INITIAL_STATE',_('INITIAL_STATE')
    # end of action
    NONE = 'NONE',_('NONE')
    


from rest_framework import serializers
from .transition import FinFlotransition
from .models import TransitionManager , workevents , workflowitems , Action
from django.conf import settings
from django.apps import apps




class Workeventsserializer(serializers.ModelSerializer):
    from_party = serializers.SlugRelatedField(read_only=True,slug_field = "description")
    to_party = serializers.SlugRelatedField(read_only=True,slug_field = "description")
    work_model = serializers.SerializerMethodField()

    class Meta:
        model = workevents
        fields = [
            'work_model',
            'id',
            'action',
            'subaction',
            'initial_state',
            'interim_state',
            'final_state',
            'event_user',
            'record_datas',
            'from_party',
            'to_party',
            'type',
            'final_value'
        ]

    def get_work_model(self,obj):
        try:
            return {"t_id" : obj.workflowitems.transitionmanager.t_id ,"model_type" : obj.workflowitems.transitionmanager.type }
        except:
            return None




class Workitemserializer(serializers.ModelSerializer):
    WorkFlowEvents = Workeventsserializer(many=True, read_only=True)
    current_from_party = serializers.SlugRelatedField(read_only=True,slug_field = "description")
    current_to_party = serializers.SlugRelatedField(read_only=True,slug_field = "description")
    work_model = serializers.SerializerMethodField()

    class Meta:
        model = workflowitems
        fields = [
            'work_model',
            'id',
            'initial_state',
            'interim_state',
            'final_state',
            'event_user',
            'action',
            'subaction',
            'next_available_transitions',
            'model_type',
            'current_from_party',
            'current_to_party',
            'WorkFlowEvents',
        ]
    

    def get_work_model(self,obj):
        try:
            return {"t_id" : obj.transitionmanager.t_id ,"model_type" : obj.transitionmanager.type }
        except:
            return None




class TransitionManagerserializer(serializers.ModelSerializer):
    workflowitems = Workitemserializer(read_only = True)
    model = serializers.SerializerMethodField()
    wf_item_id = serializers.SerializerMethodField()
    class Meta:
        model = TransitionManager
        fields = [
            'id',
            'type',
            't_id',
            'in_progress',
            'wf_item_id',
            'model',
            'workflowitems'
        ]

    def get_model(self,obj):
        try:
            arr = settings.FINFLO['WORK_MODEL']
            for i in arr:
                user = apps.get_model(i)
                qs = user.objects.filter(id = obj.t_id ).values()
                if qs.exists():
                    break
            return qs
        except:
            return None
                    

    def get_wf_item_id(self,obj):
        return obj.workflowitems.id



class Actionseriaizer(serializers.ModelSerializer):
    class Meta:
        model = Action
        fields = '__all__'



class workflowitemslistserializer(serializers.ModelSerializer):
    class Meta:
        model = workflowitems
        fields = '__all__'


class workeventslistserializer(serializers.ModelSerializer):
    class Meta:
        model = workevents
        fields = '__all__'







from django.db import models
from .middleware import get_current_user
from .validators import validate_postive
from .fields import CommaSeparatedCharField
from django.core.exceptions import ValidationError
from .validators import validate_postive
from .base_enum import Values
from django.apps import apps

try:
   from django.conf import settings
except ImportError:
    raise Exception("settings.py file is required to run this package")



## CORE CONFIGURATIONS CLASSES ##



class Flowmodel(models.Model):
    description = models.CharField(max_length=255)

    def __str__(self):
        return self.description




class PartyType(models.Model):
    description = models.CharField(max_length=255)

    def __str__(self):
        return self.description

    class Meta:
        verbose_name_plural = "1. PartyType"



class TransitionManager(models.Model):
    t_id = models.IntegerField()
    type = models.CharField(max_length = 255 )
    sub_sign = models.IntegerField(default = 0 )
    in_progress = models.BooleanField(default = False)

    # default 1 for initial submit and maker process

    def save(self, *args, **kwargs):
        self.type = self.type.capitalize()
        return super(TransitionManager,self).save(*args, **kwargs)
    
    def __str__(self):
        return "{0}_{1}".format(self.type.lower() , self.t_id)
    
    class Meta:
        verbose_name_plural = "4. TransitionManager"


class SignList(models.Model):
    sign_id = models.IntegerField(validators=[validate_postive])
    name = models.CharField(max_length = 255 , blank=True, null=True)
    sub_action_name = models.CharField(max_length = 255 , blank=True, null=True)

    def __str__(self):
        return self.name

    def save(self, *args, **kwargs):
       self.id , self.name  = self.sign_id , self.name.capitalize()
       return super(SignList, self).save(*args, **kwargs) 
    
    class Meta:
        verbose_name_plural = "4. Signatures"


    

class States(models.Model):
    description = models.CharField(max_length=255)
    
    class Meta:
        verbose_name_plural = "2. States"

    def __str__(self): 
        return self.description
    
    def save(self, *args, **kwargs):
        self.description = self.description.capitalize()
        return super(States, self).save(*args, **kwargs) 


class Action(models.Model):
    description = models.CharField(max_length=255 , blank = True , null=True , help_text = 'e.g., SUBMIT , DELETE') 
    model = models.ForeignKey(Flowmodel , on_delete = models.CASCADE , blank = True , null = True )
    from_state = models.ForeignKey(States , on_delete= models.DO_NOTHING , related_name = 'action_from_state' , blank = True , null = True , help_text = "initial from state for transition ")
    to_state = models.ForeignKey(States , on_delete= models.DO_NOTHING , blank = True , null = True , help_text = "final state for the transition to take place")
    # optional fields
    intermediator = models.BooleanField(default = False)
    from_party = models.ForeignKey(PartyType , models.DO_NOTHING , blank = True , null = True , help_text = 'this field is optional' , related_name = 'from_transition_party_type')
    to_party = models.ForeignKey(PartyType , models.DO_NOTHING , blank = True , null = True , help_text = 'this field is optional' , related_name = 'to_transition_party_type')
    stage_required = models.ForeignKey(SignList , on_delete= models.DO_NOTHING ,blank = True , null = True ,help_text = "IMPORTANT : if 0 means initial_transition ")
    sign_required = models.IntegerField(default=0,help_text = "IMPORTANT : if 0 means initial_transition " , blank=True, null=True )

 
    def save(self, *args, **kwargs):
        self.description = self.description.upper()
        try:
            sign_len = SignList.objects.get(name = self.stage_required.name )
            self.sign_required = sign_len.sign_id
        except:
            self.sign_required = 0
        return super(Action,self).save(*args, **kwargs)

    # def __str__(self):
    #     return "{0} -> {1} -> sign_required -> {2}".format(self.description , self.model, self.sign_required)
    
    def __str__(self):
        return self.description

    class Meta:
        verbose_name_plural = "3. Action"
        unique_together = ('description', 'model') 
 

class workflowitems(models.Model):
    
    created_date = models.DateTimeField(auto_now_add=True)
    transitionmanager = models.OneToOneField(TransitionManager, on_delete=models.CASCADE,blank=True, null=True )
    initial_state  = models.CharField(max_length=50,default = Values.DRAFT.value)
    interim_state = models.CharField(max_length=50,default = Values.DRAFT.value)
    final_state = models.CharField(max_length=50,default = Values.DRAFT.value) 
    event_user = models.ForeignKey(settings.AUTH_USER_MODEL, on_delete=models.DO_NOTHING , blank=True, null=True)
    next_available_transitions =  models.JSONField(blank=True, null=True)
    current_from_party = models.ForeignKey(PartyType , models.DO_NOTHING , blank = True , null = True , editable = False , related_name = 'wf_from_party')
    current_to_party = models.ForeignKey(PartyType , models.DO_NOTHING , blank = True , null = True , editable = False )
    action = models.CharField(max_length=25 , blank=True, null=True , default = Values.DRAFT.value)
    subaction = models.CharField(max_length=55 , blank=True, null=True)
    previous_action = models.CharField(max_length=55 , blank=True, null=True)
    model_type  = models.CharField(max_length=55, blank=True, null=True)
    comments = models.CharField(max_length=500,blank=True, null=True)
    is_read = models.BooleanField(default=True,blank=True, null=True)
    final_value = models.BooleanField(default=False,blank=True, null=True)
    
    class Meta:
        verbose_name_plural = "WorkFlowItem"
        ordering = ['id']

    # def save(self, *args, **kwargs):
    #    if self.from_party is None:
    #         self.from_party = None
    #    super(workflowitems, self).save(*args, **kwargs) # Call the real save() method


# WORKEVENTS
class workevents(models.Model):

    workflowitems = models.ForeignKey(workflowitems, on_delete=models.CASCADE , related_name='WorkFlowEvents')
    action = models.CharField(max_length=25, blank=True, null=True , default = Values.DRAFT.value)
    subaction = models.CharField(max_length=55 , blank=True, null=True , default = Values.DRAFT.value)
    initial_state  = models.CharField(max_length=50 , default = Values.DRAFT.value)
    interim_state = models.CharField(max_length=50,default = Values.DRAFT.value)
    final_state = models.CharField(max_length=50,default = Values.DRAFT.value)
    from_party = models.ForeignKey(PartyType , models.DO_NOTHING  , blank = True , null = True , editable = False ,related_name = 'we_from_party')
    to_party = models.ForeignKey(PartyType , models.DO_NOTHING  , blank = True , null = True , editable = False )
    event_user = models.ForeignKey(settings.AUTH_USER_MODEL,  on_delete=models.DO_NOTHING , blank=True, null=True)
    is_read = models.BooleanField(default=True,blank=True, null=True)
    record_datas = models.JSONField(blank=True, null=True)
    final_value = models.BooleanField(default=False,blank=True, null=True)
    comments = models.CharField(max_length=500,blank=True, null=True)
    created_date = models.DateTimeField(auto_now_add=True)
    type = models.CharField(max_length=55, blank=True, null=True)
    

    class Meta:
        verbose_name_plural = "WorkFlowEvent"
        ordering = ['id']


    # def save(self , *args, **kwargs):
    #     try:
    #         work_model = settings.FINFLO['WORK_MODEL']  
    #         for iter in work_model:
    #             gets_model =  apps.get_model(iter)
    #             values = gets_model.objects.filter(id = self.workflowitems.transitionmanager.t_id).value_list()
    #             print("the data is " ,values)
    #             if values.exists():
    #                 break
    #             continue
    #         self.record_datas = {"values" : values}
    #         print(self.record_datas)
    #         return super(workevents, self).save( *args, **kwargs)
    #     except :
    #         self.record_datas = {"values" : values}
    #         return super(workevents, self).save( *args, **kwargs)
    
        








from django.core.exceptions import ValidationError
from django.utils.translation import gettext_lazy as _

def validate_postive(value):
    if value < 0:
        raise ValidationError(
            _('%(value)s is not an positive number'),
            params={'value': value},
        )


from django.contrib import admin
from .models import Action, PartyType, States, TransitionManager , workevents , workflowitems , SignList 
# Register your models here.
from django.contrib import messages
from django.utils.translation import ngettext
from django.conf import settings


class TransitionModeladmin(admin.ModelAdmin):
    list_display = ("t_id", "type", "in_progress")

    def change_to_draft(self, request, queryset):
        updated = queryset.update(in_progess = False )
        self.message_user(request, ngettext(
            '%d story was successfully marked as published.',
            '%d stories were successfully marked as published.',
            updated,
        ) % updated, messages.SUCCESS)

class CustomActionAdminModel(admin.ModelAdmin):

    def has_delete_permission(self, request, obj=None):
        return obj is None or obj.pk != 1


class customadminforsign(admin.ModelAdmin):

    def has_delete_permission(self, request, obj=None):
        return obj is None or obj.pk != 0
    

admin.site.register(TransitionManager,TransitionModeladmin)
admin.site.register(Action , CustomActionAdminModel )
admin.site.register(States)
admin.site.register(PartyType)
admin.site.register(SignList , customadminforsign)
admin.site.register(workflowitems)
admin.site.register(workevents)


# Generated by Django 3.2.5 on 2022-10-14 10:50

from django.conf import settings
from django.db import migrations





# RUN ON MIGRATIONS  -- remigrate if work_model is updated 

def update_flow_model(apps, schema_editor):
    flowmodel = apps.get_model("finflo.Flowmodel")
    try:
        a = settings.FINFLO['WORK_MODEL']
        db_alias = schema_editor.connection.alias
        for i in a:
            flowmodel.objects.using(db_alias).update_or_create(description = i)
    except:
        raise LookupError("Couldn't find work_model in settings.py check documentation")


class Migration(migrations.Migration):

    dependencies = [
        ('finflo', '0001_initial'),
    ]

    operations = [
        migrations.RunPython(update_flow_model)
    ]



#  base_enum 
# while stage = False:
    # v = 1
    # try:
    #     for item in itens:
    #         b = itens[1 + v]
    #         print(b)
    #         c = b
    #         if b == itens[-1]:
    #             break
    #     print("the value is ", c)
    # except:
    #     print("cant do this ")




 # trans_length = int(stage) - int(gets_model.sign_required)
                        # if trans_length != 1 and int(stage) == int(gets_sign):
                        #     trans_length_var = trans_length - 1
                        # else:
                        # while stage == True:
                        #     try:
                        #             for item in sign_list:
                        #                 b = sign_list[1 + trans_length_var]
                        #                 print(b)
                        #                 c = b
                        #                 if b == sign_list[-1]:
                        #                     break
                        #             print("the value is ", c)
                        #     except:
                        #             print("cant do this ")
                        #     return None


from django import dispatch
from django.db.models.signals import post_save , pre_save , post_delete

from .middleware import get_current_user
from .models import States, TransitionManager, workevents, workflowitems 
from django.conf import settings
from .transition import FinFlotransition



# CUSTOM MULTIPLE RECEIVER SIGNALS 


def finflo_receiver(signal, senders ,  **kwargs):
    def decorator(receiver_func):
        for sender in senders:
            if isinstance(signal, (list, tuple)):
                for s in signal:
                    s.connect(receiver_func, sender=sender, **kwargs)
            else:
                signal.connect(receiver_func, sender=sender,**kwargs)

        return receiver_func

    return decorator

# 



# MAIN SIGNAL RECEIVER FOR TRANSITION HANDLING - 4/8/2022 by anand


a = [ite for ite in settings.FINFLO['WORK_MODEL']]
@finflo_receiver(post_save, senders = a)
def create(sender, instance, **kwargs):
        a = str(sender)
        remove_characters = ["class", "models.","'","<",">"," "]
        for i in remove_characters:
            a = a.replace(i,"")
        manager = TransitionManager.objects.create(
            type = a.lower() , t_id = instance.id
        )
        gets_transition = FinFlotransition(t_id = manager.id , type =   a.lower() , action  = None,interim = None , source = None , target = None)
        states = States.objects.get(id = 1)
        wf = workflowitems.objects.create(transitionmanager = manager ,model_type = a.upper() , event_user = get_current_user() , initial_state = states.description ,interim_state = states.description, final_state = states.description)
        we = workevents.objects.create(workflowitems = wf , type = a.upper() , event_user = get_current_user() , record_datas = gets_transition.get_record_datas(), initial_state = states.description ,interim_state = states.description,  final_state = states.description)
        wf.save()
        we.save()
        manager.save()




import json
from .models import (
    SignList,
    TransitionManager,
    Action,
    workevents,
    workflowitems,
    Flowmodel
)
from django.core import serializers as core_serializers
from .middleware import get_current_user
from .exception import (
    ModelNotFound,
    TransitionNotAllowed,
    ReturnModelNotFound,
)
from django.db.models import Q
from django.conf import settings
from collections import deque
from django.apps import apps
from django.forms.models import model_to_dict




####################################################
#############       CORE     #######################
####################################################


class FinFlotransition():

    # BASE CONSTRUCTORS # 
    
    def __init__(self, type, t_id , action = None ,source = None,interim = None, target = None, from_party = None, to_party = None):
        # typing support needed 
        self.type = type 
        self.t_id = t_id 
        self.action = action if action else None
        self.source = source if source else None
        self.interim = interim if interim else None
        self.target = target if target else None
        self.from_party = from_party if from_party else None
        self.to_party = to_party if to_party else None
        #  conditions
        gets_current_Action = None
        if action:
            gets_action = self.gets_base_action()
            base_model = self.gets_base_model()
            gets_return = self.gets_default_return_action()
            if gets_action.intermediator == True :
                self.intermediator_transition_handler(type = self.type, action = self.action , t_id = self.t_id)
            elif gets_return.description == action.upper():
                self.return_transition()
            else :
                self.transition()
        else:
            self.manualtransitions()
        # return None

    def __repr__(self):
        return "the id is %s and type is %s" % (self.t_id, self.type)

    
    def __str__(self):
        return "the id is %s and type is %s" % (self.t_id, self.type)


    # GETS THE ALL MODEL VALUE FOR TRANSITION #

    def gets_base_action(self):
        try:
            gets_action_data = Action.objects.get(description=self.action.upper(), model__description = self.type.lower() )
            return gets_action_data
        except:
            raise ModelNotFound()
    
    def gets_all_models(self):
        try:
            gets_model = TransitionManager.objects.get(type = self.type.upper() , t_id = self.t_id)
            gets_flows = Flowmodel.objects.get(description  = self.type.lower())
            gets_action = Action.objects.get(Q(model = gets_flows.id ) | Q(model = None), description=self.action)
            gets_wf  = FinFlotransition.gets_wf_item(gets_model)
            sign_lists = sub_action_list = []
            try:
                for item in SignList.objects.all():
                    sign_lists.append(item.name)
                    if item.name == gets_action.stage_required.name :
                        break
                next_avail_trans = sign_lists[gets_model.sub_sign : ]
                next_avail_trans_value = deque(next_avail_trans)
                next_avail_trans_value.popleft()
                next_states = list(next_avail_trans_value)
                next_avail_Transition = {'values' : next_states}
            except:
                next_avail_Transition = None
                pass
            return gets_model , gets_action , gets_flows , gets_wf , sign_lists , next_avail_Transition
        except:
            raise ModelNotFound()

    # GETS TRANSITION MANAGER

    def gets_base_model(self):
        try:
            gets_model = TransitionManager.objects.get(type = self.type.upper() , t_id = self.t_id)
            return gets_model 
        except:
            ModelNotFound()


    # def get_action(self):
    #     self._action = None

    def get_record_datas(self):
        overall_model = FinFlotransition.gets_base_model(self)
        try:
            work_model = settings.FINFLO['WORK_MODEL']  
            for iter in work_model:
                gets_model =  apps.get_model(iter)
                query_data = gets_model.objects.filter(id = overall_model.workflowitems.transitionmanager.t_id)
                base_serialized_Data = core_serializers.serialize('json', query_data)
                if query_data.exists():
                    break
                continue
            return {"values" : json.loads(base_serialized_Data)}
        except:
            return {"values" : None}
           

    # action = property(get_action, set_action)

    

    # MANUAL TRANSITION WITH SOURCE , INTERIM , AND TARGET STATES

    def manualtransitions(self):
        try:
            queryset = TransitionManager.objects.get(t_id = self.t_id , type = self.type)
            wf = workflowitems.objects.update_or_create( transitionmanager= queryset , defaults= {"initial_state" : self.source, "interim_state" : self.interim ,
                    "final_state" : self.target, "model_type" : self.type.upper(), "event_user" : get_current_user() , "current_from_party " : self.from_party, "current_to_party " : self.to_party})
            wf.save()
            workevents.objects.create(workflowitems = wf , event_user=get_current_user(),  initial_state=self.source  , 
                    interim_state = self.interim , final_state=self.target ,type=self.type.upper() , record_datas = self.get_record_datas() , from_party = self.from_party , to_party = self.to_party)
        except:
            return None  
       



    # GETS THE DEFAULT RETURN ACTION #

    def gets_default_return_action(self):
        try:
            qs = Action.objects.get(id = 1)
            return qs
        except:
            raise ReturnModelNotFound()


    # VALUE FOR SIGN_LIST #

    def get_value(sign_list, i):
        try:
            return sign_list[i]
        except IndexError:
            return 0



    # GETS WORKFLOW MODEL ID #

    def gets_wf_item(gets_model):
        ws = workflowitems.objects.get(transitionmanager=gets_model.id)
        return ws
        
    
    

    # SPECIAL FUNCTION FOR RETURN TRANSITION  #

    def return_transition(self):
        
        overall_model = FinFlotransition.gets_all_models(self)
       
        if self.action == self.gets_default_return_action().description.upper():
            wf = workflowitems.objects.update_or_create( transitionmanager=overall_model[0] or overall_model[0].id, defaults= {"initial_state" : overall_model[3].initial_state, "interim_state" : overall_model[1].to_state.description ,
                    "final_state" : overall_model[1].to_state.description, "action" : self.action, "subaction" : self.action , "next_available_transitions" : None, "model_type" : self.type.upper(), "event_user" : get_current_user() , "current_from_party" : overall_model[1].from_party , "current_to_party" : overall_model[1].to_party})
            workevents.objects.create(workflowitems=overall_model[3] , event_user=get_current_user(),  initial_state=overall_model[3].initial_state,final_value = True , record_datas = self.get_record_datas() , 
                interim_state = overall_model[1].to_state.description , final_state=overall_model[1].to_state.description, action= self.action, subaction = self.action ,type=self.type.upper(), from_party = overall_model[1].from_party , to_party = overall_model[1].to_party)
            overall_model[0].sub_sign = 0
            overall_model[0].save()                      



    # INTERIM ACTION FOR THE TRANSITION HANDLING #

    def intermediator_transition_handler(self, type, action, t_id):
        overall_model = FinFlotransition.gets_all_models(type = type, action = action , t_id = t_id)
        
        # len action and sub_sign
        
        if overall_model.gets_model.sub_sign <= overall_model.gets_action.sign_required:
            def Transition_Handler():
                    gets_sign = overall_model.gets_action.sign_required
                    
                    
                    # if len(overall_model.sign_lists)-1 != overall_model.gets_model.sub_sign:
                    #     ws = workflowitems.objects.update_or_create( transitionmanager=overall_model.gets_model or overall_model.gets_model.id, defaults= {"initial_state" : gets_action.from_state.description, "interim_state" :  sign_lists[1 + gets_model.sub_sign], 
                    #         "final_state" : overall_model.gets_action.to_state.description, "action" : action, "subaction" : overall_model.sign_lists[1 + gets_model.sub_sign], "model_type" : type.upper(), "event_user" : get_current_user() , "current_from_party" : gets_action.from_party , "current_to_party" : gets_action.to_party})
                    #     we = workevents.objects.create(workflowitems=overall_model.gets_wf, event_user=get_current_user(),  initial_state=overall_model.gets_action.from_state.description,
                    #                               interim_state = overall_model.sign_lists[1 + overall_model.gets_model.sub_sign], final_state=overall_model.gets_action.to_state.description, action=action, subaction=sub_action[0], type=type.upper(), from_party = gets_action.from_party , to_party = gets_action.to_party)
                    #     overall_model.gets_model.sub_sign += 1
                    #     overall_model.gets_model.save()

                    # elif len(overall_model.sign_lists)-1 == int(overall_model.gets_model.sub_sign):
                    #     ws = workflowitems.objects.update_or_create( transitionmanager=overall_model.gets_model or gets_model.id, defaults= {"initial_state" : gets_action.from_state.description, "interim_state" : gets_action.to_state.description ,
                    #         "final_state" : overall_model.gets_action.to_state.description, "action" : action, "subaction" : sign_lists[1 + gets_model.sub_sign], "model_type" : type.upper(), "event_user" : get_current_user() , "current_from_party" : gets_action.from_party , "current_to_party" : gets_action.to_party})
                    #     workevents.objects.create(workflowitems=gets_wf, event_user=get_current_user(),  initial_state=gets_action.from_state.description,final_value = True , 
                    #                               interim_state = gets_action.to_state.description, final_state=gets_action.to_state.description, action=action, subaction=sub_action[0], type=type.upper(), from_party = gets_action.from_party , to_party = gets_action.to_party)
                    #     overall_model.gets_model.sub_sign += 1
                    #     overall_model.gets_model.save()
            return Transition_Handler()   
            
        else:
            raise TransitionNotAllowed()
        


    ## CORE TRANSITION ###

    def transition(self):
       
        overall_model = FinFlotransition.gets_all_models(self)
        
        # if ((gets_model.sub_sign <= gets_action.sign_required) and (gets_model.sub_sign != gets_action.sign_required)):
        if overall_model[0] is not None :
            def Transition_Handler():
        
                    # NO SIGN -> ONLY_MAKER TRANSITION 
                    
                    # if len(overall_model[4])  == 0:
                    #     # print("SET 1 ")
                    #     workflowitems.objects.update_or_create( transitionmanager=overall_model[0] or overall_model[0].id, defaults= {"initial_state" : overall_model[1].from_state.description, "interim_state" :  overall_model[1].from_state.description, 
                    #         "final_state" : overall_model[1].to_state.description, "action" : self.action, "subaction" : None, "model_type" : self.type.upper(), "event_user" : get_current_user() , "current_from_party" : overall_model[1].from_party , "current_to_party" : overall_model[1].to_party})
                    #     workevents.objects.create(workflowitems=overall_model[3], event_user=get_current_user(),  initial_state=overall_model[1].from_state.description,
                    #                               interim_state = overall_model[1].from_state.description, final_state=overall_model[1].to_state.description, action= self.action, subaction= None , type= self.type.upper(), from_party = overall_model[1].from_party , to_party = overall_model[1].to_party)
                    #     overall_model[0].sub_sign = 0
                    #     overall_model[0].save()

                    # # MAKER TRANSITION 

                    # elif overall_model[0].sub_sign == 0:
                    #     # print("SET 2 ")
                    #     ws = workflowitems.objects.update_or_create( transitionmanager=overall_model[0] or overall_model[0].id, defaults= {"initial_state" : overall_model[1].from_state.description, "interim_state" :  overall_model[4][1 + overall_model[0].sub_sign], 
                    #         "final_state" : overall_model[1].to_state.description, "next_available_transitions" : overall_model[5],"action" : self.action, "subaction" : overall_model[4][overall_model[0].sub_sign], "model_type" : self.type.upper(), "event_user" : get_current_user() , "current_from_party" : overall_model[1].from_party , "current_to_party" : overall_model[1].from_party})
                    #     workevents.objects.create(workflowitems=overall_model[3], event_user=get_current_user(),  initial_state=overall_model[1].from_state.description,
                    #                               interim_state = overall_model[4][1 + overall_model[0].sub_sign], record_datas = self.get_record_datas() , final_state=overall_model[1].to_state.description, action= self.action, subaction= self.action, type= self.type.upper(), from_party = overall_model[1].from_party , to_party = overall_model[1].from_party)
                    #     overall_model[0].sub_sign +=1
                    #     overall_model[0].save()

                    #  TRANSITION 

                    if len(overall_model[4]) != overall_model[0].sub_sign:
                        # print("SET 3 ")
                        # print(overall_model[0].sub_sign)
                        try:
                            ws = workflowitems.objects.update_or_create( transitionmanager=overall_model[0] or overall_model[0].id, defaults= {"initial_state" : overall_model[1].from_state.description, "interim_state" :  overall_model[4][1 + overall_model[0].sub_sign], 
                                "final_state" : overall_model[1].to_state.description, "next_available_transitions" : overall_model[5],"action" : self.action, "subaction" : overall_model[4][overall_model[0].sub_sign], "model_type" : self.type.upper(), "event_user" : get_current_user() , "current_from_party" : overall_model[1].from_party , "current_to_party" : overall_model[1].from_party})
                            workevents.objects.create(workflowitems=overall_model[3], event_user=get_current_user(),  initial_state=overall_model[1].from_state.description,
                                                    interim_state = overall_model[4][1 + overall_model[0].sub_sign], record_datas = self.get_record_datas() ,final_state=overall_model[1].to_state.description, action= self.action, subaction= self.action, type= self.type.upper(), from_party = overall_model[1].from_party , to_party = overall_model[1].from_party)
                            overall_model[0].sub_sign += 1 
                            overall_model[0].in_progress = True
                            overall_model[0].save()
                        except:
                            ws = workflowitems.objects.update_or_create( transitionmanager=overall_model[0] or overall_model[0].id, defaults= {"initial_state" : overall_model[1].from_state.description, "interim_state" :  overall_model[1].to_state.description, 
                            "final_state" : overall_model[1].to_state.description, "action" : self.action,  "next_available_transitions" : None , "subaction" : self.action, "model_type" : self.type.upper(), "event_user" : get_current_user() , "current_from_party" : overall_model[1].from_party , "current_to_party" : overall_model[1].to_party , "final_value" : True})
                            workevents.objects.create(workflowitems=overall_model[3], event_user=get_current_user(),  initial_state=overall_model[1].from_state.description,
                                                  interim_state = overall_model[1].to_state.description,record_datas = self.get_record_datas() ,  final_state=overall_model[1].to_state.description, action= self.action, subaction= self.action, type = self.type.upper(), from_party = overall_model[1].from_party , to_party = overall_model[1].to_party , final_value = True)
                            overall_model[0].sub_sign = 0
                            overall_model[0].in_progress = False
                            overall_model[0].save()

            return Transition_Handler()   
        else:
            raise TransitionNotAllowed()




## BASIC EXCEPTION CLASSES 


from rest_framework.exceptions import APIException

class ModelNotFound(APIException):
    status_code = 204
    default_detail = 'Matching action and type not found , try with some other type '
    default_code = 'model_object_not_found'

class VersionError(Exception):
    pass

class TransitionNotAllowed(APIException):
    status_code = 204
    default_detail = 'Transition Not Allowed .'
    default_code = 'transition_unavailable'


class ReturnModelNotFound(APIException):
    status_code = 204
    default_detail = 'Unable to find the default model object  , try again .'
    default_code = 'model_object_not_found'



class SignaturesNotMatching(APIException):
    status_code = 204
    default_detail = 'Signatures not matching , only positive index accepted'
    default_code = 'signatures_index_not_matching'


from enum import Enum

class Values(Enum):
    DRAFT = "DRAFT"
    RETURN = "RETURN"
    MAKER = "MAKER"
    AWAITING_SIGN_A = "AWAITING_SIGN_A"
    AWAITING_SIGN_B = "AWAITING_SIGN_B"
    AWAITING_SIGN_C = "AWAITING_SIGN_C"
    SIGN_A  = "SIGN_A"
    SIGN_B = "SIGN_B"
    SIGN_C = "SIGN_C"



# IMP_NOTE:  THE DRAFT IS DEFAULT VALUE FOR A TRANSITION STARTED PACK 


from .models import Action, SignList
from rest_framework.views import APIView
from rest_framework import status
from rest_framework.response import Response
from rest_framework.permissions import IsAuthenticated
from .transition import FinFlotransition
from rest_framework.generics import ListAPIView , ListCreateAPIView 
from django.conf import settings
from .serializer import  (
    TransitionManagerserializer,
    Actionseriaizer,
    workflowitemslistserializer,
    workeventslistserializer,
    Workitemserializer
)
from .models import (
    Action, 
    TransitionManager, 
    workevents, 
    workflowitems
)
from django.db.models import Q



####################################################
################      API       ####################
####################################################



# 1 . TRANSITION MAKING API


class TransitionApiView(APIView):
    permission_classes = [IsAuthenticated]
    
    def post(self,request):
        type = request.data.get("type")
        action = request.data.get("action")
        t_id = request.data.get("t_id")
        # extra datas for manual transitions
        source = request.data.get("source")
        interim = request.data.get("interim")
        target = request.data.get("target")

        if type and t_id  is not None:
            transitions = FinFlotransition(action = action , type = type , t_id = t_id , source = source , interim = interim ,target = target )
            return Response({"status" : "Transition success"},status = status.HTTP_200_OK)
        return Response({"status" : "Transition failure"},status = status.HTTP_204_NO_CONTENT)


    


#  2 . ALL WORK_MODEL LIST 


class DetailsListApiView(ListAPIView):
    queryset = TransitionManager.objects.all()
    serializer_class = TransitionManagerserializer
    permission_classes = [IsAuthenticated]
    
    def get_queryset(self):
        type = self.request.query_params.get('type',None)
        t_id = self.request.query_params.get('t_id',None)
        if t_id and type is not None :
            queryset = TransitionManager.objects.filter(Q(type__icontains = type ) | Q(t_id = int(t_id)))
        elif type is not None and t_id is None:
            queryset = TransitionManager.objects.filter(type__icontains = type)
        else:
            queryset = TransitionManager.objects.all()
        return queryset

    def list(self, request):
        queryset = self.get_queryset()
        serializer = TransitionManagerserializer(queryset, many=True)
        return Response({"status": "success", "type" : settings.FINFLO['WORK_MODEL'] , "data": serializer.data}, status=status.HTTP_200_OK)



# 3 . WORFLOW API 

class WorkFlowitemsListApi(ListAPIView):
    queryset = workflowitems.objects.all()
    serializer_class = Workitemserializer
    permission_classes = [IsAuthenticated]

    def list(self, request):
        queryset = workflowitems.objects.all()
        serializer = Workitemserializer(queryset, many=True)
        return Response({"status": "success", "data": serializer.data}, status=status.HTTP_200_OK)




# WORKEVENTS API 


class WorkEventsListApi(ListAPIView):
    queryset = workevents.objects.all()
    serializer_class = workeventslistserializer
    permission_classes = [IsAuthenticated]

    def list(self, request):
        queryset = workevents.objects.all()
        serializer = workeventslistserializer(queryset, many=True)
        return Response({"status": "success", "data": serializer.data}, status=status.HTTP_200_OK)



# ACTION CREATE AND LIST API 


class ActionListApi(ListCreateAPIView):
    queryset = Action.objects.all()
    serializer_class = Actionseriaizer
    permission_classes = [IsAuthenticated]

    def list(self, request):
        queryset = Action.objects.all()
        serializer = Actionseriaizer(queryset, many=True)
        return Response({"status": "success" , "data": serializer.data}, status=status.HTTP_200_OK)


    def post(self, request):
        serializer = Actionseriaizer(data=request.data)
        if serializer.is_valid():
            serializer.save()
            return Response({"status": "success", "data": serializer.data}, status=status.HTTP_201_CREATED)
        return Response({"status": "failure", "data": serializer.errors},status=status.HTTP_204_NO_CONTENT)








from django import forms
from django.core import validators
from django.core.exceptions import ValidationError

class MinLengthValidator(validators.MinLengthValidator):
    message = 'Ensure this value has at least %(limit_value)d elements (it has %(show_value)d).'

class MaxLengthValidator(validators.MaxLengthValidator):
    message = 'Ensure this value has at most %(limit_value)d elements (it has %(show_value)d).'


class CommaSeparatedCharField(forms.Field):
    def __init__(self, dedup=True, max_length=None, min_length=None, *args, **kwargs):
        self.dedup, self.max_length, self.min_length = dedup, max_length, min_length
        super(CommaSeparatedCharField, self).__init__(*args, **kwargs)
        if min_length is not None:
            self.validators.append(MinLengthValidator(min_length))
        if max_length is not None:
            self.validators.append(MaxLengthValidator(max_length))

    def to_python(self, value):
        if value in validators.EMPTY_VALUES:
            return []

        value = [item.strip() for item in value.split(',') if item.strip()]
        if self.dedup:
            value = list(set(value))

        return value

    def clean(self, value):
        value = self.to_python(value)
        self.validate(value)
        self.run_validators(value)
        return value
